require 'net/http'
require 'hpricot'
require 'open-uri'
require 'uri'
require 'pp'
require 'ferret'

require 'Models/cached_queries'

class QueryRewriter
  def initialize(query)
    init_query(query)
  end
  
  def init_query(query)
    @query = query
    puts query
    Ferret::Index::Index.new(:path => 'e:/wiki/index', :analyzer => Ferret::Analysis::WhiteSpaceAnalyzer.new) do |index|
      hits = []
      index.search_each("synonyms:\"SYN #{query.downcase} SYN\"", :limit => 100, :sort => 'weight DESC') do |id, score|
        hits << id
      end

      pp hits
      
      if hits.length == 1
        id = hits.first
        @synonyms = []
        index[id][:synonyms].scan(/SYN (.*) SYN (\d+)/) do |match|
          @synonyms << {:name => match[0], :weight => match[1].to_i}
        end
      elsif hits.length >= 2
        @entities = []
        hits.each do |id| 
          hit = index[id]
          @entities << {:name => hit[:name], :weight => hit[:weight]}
        end
      end
      
      #      puts "id: #{id} score: #{score}\n"
      #      puts index[id][:name]
      #    
      #      entities = []
      #      index[id][:synonyms].scan(/SYN (.*) SYN (\d+)/) do |match|
      #        entities << [match[0], match[1].to_i]
      #      end
      #      entities.sort { |x, y| y[1] <=> x[1] }.each do |entity, weight|
      #        puts "#{entity} --- #{weight}"
      #      end
      #    
      #      puts index[id][:spec]
      #      #    puts index[id][:synonyms]
      #      puts "\n\n"

    end
  end
  
  def entities; @entities end
  
  def synonym_queries
    #['george bush', 'bush', 'president bush', 'george w. bush']
    #['"united nations"', 'un']
    unless @syns
      @syns = @synonyms.sort { |x, y| y[:weight] <=> x[:weight] }
      @syns = @syns[0..4] if @syns.length > 5
      @syns.collect! { |s| s[:name] }
    end
    @syns
  end
  
  def expanded_query
    "#{@query} OR " + synonym_queries.collect { |s| "\"#{s}\"" }.join(' OR ')
    #"(\"united nations\") OR (un)"
    #    "(president bush) or (bush)"
    #    synonym_queries.collect { |q| "(#{q})"} .join('  ')
  end
end

module SearchHelper
  def host
    "news.search.yahoo.com"
  end
  
  def port
    80
  end
  
  def query_to_querystring(query)
    "/news/search?p=#{CGI::escape(query)}&c=&n=20"
  end
  
  def extract(doc, expr)
    list = []
    doc.search(expr) { |e| list << e }
    list
  end
  
  def get_from_cache(query)
    cached_query = CachedQuery.find(:first, :conditions => ["uri = ?", query])
    cached_query.body if cached_query
  end
  
  def store_in_cache(query, body)
    cq = CachedQuery.new(:uri => query, :body => body)
    cq.save
  end
  
  def fetch_queries(queries)
    results = []
    
    pp queries
    
    http = nil
    begin
      queries.each do |query|
        body = get_from_cache(query)
        
        unless body
          http ||= Net::HTTP.start(host, port)
          request = Net::HTTP::Post.new(query_to_querystring(query))
          request.add_field 'Referer', 'http://news.yahoo.com/'
          response = http.request(request)
          body = response.body

          store_in_cache(query, body)
        end
        
        doc = Hpricot(body)
        hits = [extract(doc, "a[@class=yschttl]"), extract(doc, "div[@class=yschabstr]")].transpose.collect do |e|
          title, desc = e
          href = title[:onmouseover]
          href = href["return window.status='".length..href.length - 2]
          #          puts title[:onmouseover]
          #          puts href
          {:desc => desc.inner_text.strip, :href => href, :title => title.inner_text.strip}
        end
        
        puts "Query: #{query} Hits: #{hits.length}"
        hits.each do |hit|
          puts "#{hit[:href]} - #{hit[:title]}"
        end
        puts "\n\n"
        
        results << {
          #          :body => body,
          :query => query,
          :hits => hits }
      end
    ensure
      http.finish if http
    end
    
    return results
  end
  
  def expanded_query; @expanded_query end
  def entities; @entities end
  
  def run_query(query)
    rewriter = QueryRewriter.new(query)
    @entities = rewriter.entities
    
    unless @entities
      results = fetch_queries([rewriter.expanded_query] + rewriter.synonym_queries.collect { |s| "\"#{s}\"" })
    
      @expanded_query = rewriter.expanded_query

      hit_rank = {}
      results[1..results.length].each do |result|
        result[:hits].each_with_index do |hit, index|
          hr = {:query => result[:query], :index => index + 1}
          hit_rank[hit[:href]] = hit_rank.fetch(hit[:href], []) + [hr]
          hit_rank[hit[:title]] = hit_rank.fetch(hit[:title], []) + [hr]
        end
      end

      #    pp hit_rank

      hits = results[0][:hits].collect do |hit|
        s1 = hit_rank[hit[:href]]
        s2 = hit_rank[hit[:title]]
        s = []
        s += s1 if s1
        s += s2 if s2
        s.uniq!
        
        {:synonyms => s}.merge(hit)
      end
      qhits = results[1..results.length]
      [hits, qhits]
    end
  end
end
