// Sorter.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

const unsigned int BUFFER_SIZE = 512 * 1024 * 1024;
const unsigned int CHUNK_BUFFER_SIZE = 4 * 1024 * 1024;
const std::string HEADER("TUPLESTREAM\x22\x2a\x6f\xf7");

enum FieldType { INTEGER, STRING };
const int IntegerSize = 4;

typedef std::vector<size_t> IndexList;

class TupleStreamDescription {
public:
	TupleStreamDescription(std::ifstream& input, IndexList comparatorFields): comparatorFields_(comparatorFields), stringFieldCount_(0) {
		readHeader(input);
//		compareFields_.push_back(0);
//		compareFields_.push_back(1);
//		compareFields_.push_back(2);
//		compareFields_.push_back(3);
	}

	int compareTuples(const char* x, const char* y) {
		IndexList::const_iterator it = comparatorFields_.begin(), end = comparatorFields_.end();
		for (; it != end; ++it) {
			size_t fieldNum = *it;

//			int fieldIndex = fields_[fieldNum] == INTEGER ? 

			long long value1 = fieldToUInt(x, fieldNum);
			long long value2 = fieldToUInt(y, fieldNum);

			long long c;
			switch (fields_[fieldNum]) {
			case INTEGER:
				c = value1 - value2;
				break;
			case STRING:
				{
					const char* str1 = fieldToString(x, fieldNum),* str2 = fieldToString(y, fieldNum);
					c = strncmp(str1, str2, static_cast<unsigned int>(std::min(value1, value2)));
					long long c2 = c;
					if (c == 0) {
						c = value1 - value2;
					}

//					if (c2 == 0 && value1 != value2) {
//						std::cout << std::string(str1, value1) << "|||" << std::string(str2, value2) << "|||" << c << std::endl;
//					}
				}
				break;
			default:
				assert(false);
			}

			if (c != 0) {
				return c < 0 ? -1 : 1;
			}
		}

		return 0;
	}

	size_t fieldCount() {
		return fields_.size();
	}

	inline size_t staticLength() {
		return fields_.size() * IntegerSize;
	}

	size_t tupleLength(char* data) {
		size_t length = staticLength();
		for (size_t i = 0; i < stringFieldCount_; ++i) {
			length += ptrToUInt(data);
			data += IntegerSize;
		}
		return length;
	}

	void print(const char* tupleData) {
		IndexList::const_iterator it = comparatorFields_.begin(), end = comparatorFields_.end();
		for (; it != end; ++it) {
			size_t fieldNum = *it;

			long long value = fieldToUInt(tupleData, fieldNum);

			switch (fields_[fieldNum]) {
			case INTEGER:
				std::cout << value << " ";
				break;
			case STRING:
				{
					const char* str = fieldToString(tupleData, fieldNum);
					std::cout << "'" << std::string(str, value) << "' ";
				}
				break;
			default:
				assert(false);
			}
		}
		std::cout << std::endl;
	}

	void writeHeader(std::ostream& output) {
		output.write(&HEADER[0], HEADER.size());
		writeUInt(output, fields_.size());

		std::vector<FieldType>::const_iterator it = fields_.begin(), end = fields_.end();
		for (; it != end; ++it) {
			writeUInt(output, static_cast<char>(*it));
		}
	}
private:
	template<typename T> inline const T castPtrToValue(const char* data) {
		return *reinterpret_cast<const T*>(data);
	}

	inline unsigned int ptrToUInt(const char* data) {
		return castPtrToValue<unsigned int>(data);
	}

	inline unsigned int fieldToUInt(const char* tupleData, unsigned int fieldNum) {
		switch (fields_[fieldNum]) {
			case INTEGER:
				return ptrToUInt(fieldToString(tupleData, fieldNum));
			case STRING:
				{
					int strIndex = 0;
					for (size_t i = 0; i < fieldNum; ++i) {
						if (fields_[i] == STRING) {
							++strIndex;
						}
					}
					return ptrToUInt(tupleData + strIndex * IntegerSize);
				}
			default:
				assert(false);
				return 0;
		}
		//return castPtrToValue<unsigned int>(tupleData + fieldNum * 4);
		//		return castPtrToValue<unsigned int>(tupleData + fieldOffset_[fieldNum]);
	}

	const char* fieldToString(const char* tupleData, size_t fieldNum) {
		size_t offset = stringFieldCount_ * IntegerSize;
		int str_index = 0;
		for (size_t i = 0; i < fieldNum; ++i) {
			switch (fields_[i]) {
			case INTEGER:
				offset += IntegerSize;
				break;
			case STRING:
				offset += ptrToUInt(tupleData + (4 * str_index++));
				break;
			default:
				assert(false);
			}
		}
		return tupleData + offset;
	}

	void readHeader(std::ifstream& input) {
		std::vector<char> buffer(HEADER.size(), 0);
		input.read(&buffer[0], HEADER.size());
		size_t bytesRead = input.gcount();

		if (bytesRead != HEADER.size() || !std::equal(HEADER.begin(), HEADER.end(), buffer.begin())) {
			throw std::exception("Not a tuple stream file.");
		}

		unsigned int fieldCount;
		readUInt(input, fieldCount);

		for (size_t i = 0; i < fieldCount; ++i) {
			unsigned int field;
			readUInt(input, field);
			fields_.push_back(static_cast<FieldType>(field));

			if (field == STRING) {
				++stringFieldCount_;
			}
		}
	}

	void readUInt(std::istream& input, unsigned int& value) {
		input.read(reinterpret_cast<char*>(&value), sizeof(value));
		size_t bytesRead = input.gcount();
		if (bytesRead != sizeof(value)) {
			throw std::exception("Expected to read an unsigned int, but not enough data.");
		}
	}

	void writeUInt(std::ostream& output, unsigned int value) {
		output.write(reinterpret_cast<char*>(&value), sizeof(value));
	}

	IndexList comparatorFields_;
	std::vector<FieldType> fields_;
	std::vector<int> fieldOffset_;
	size_t stringFieldCount_;
};

struct TupleComparator {
	TupleComparator(TupleStreamDescription& tsd, std::vector<char>& dataBuffer): tsd_(tsd), dataBuffer_(dataBuffer) { }

	bool operator()(unsigned int x, unsigned int y) const {
		return tsd_.compareTuples(&dataBuffer_[x], &dataBuffer_[y]) <= 0;
	}
private:
	std::vector<char>& dataBuffer_;
	TupleStreamDescription& tsd_;
};

class TupleStreamSorter {
public:
	TupleStreamSorter(const std::string input, const std::string output, IndexList comparatorFields): input_(input), output_(output), comparatorFields_(comparatorFields), dataBuffer_(BUFFER_SIZE), dataBufferLen_(0) {
		//fields_.push_back(INTEGER);
		//fields_.push_back(STRING);
		//fields_.push_back(INTEGER);
		//fields_.push_back(STRING);
		//fields_.push_back(STRING);
		//fields_.push_back(STRING);
		//fields_.push_back(INTEGER);
		//		comparatorFields_.push_back(0);
		//		comparatorFields_.push_back(1);

		//		staticLength_ = fields_.size() * 4;

		//stringFields_ = 0;
		//for (size_t i = 0; i < fields_.size(); ++i) {
		//	if (fields_[i] == STRING) {
		//		++stringFields_;
		//	}
		//}
	}

	void run() {
		inStream_.open(input_.c_str(), std::ios_base::in | std::ios_base::binary);
		TupleStreamDescription tsd(inStream_, comparatorFields_);

		std::vector<std::string> chunkFileNames;

//		if (false) {
		while (true) {
			inStream_.read(&dataBuffer_[dataBufferLen_], dataBuffer_.size() - dataBufferLen_);
			size_t bytesRead = inStream_.gcount();
			if (bytesRead <= 0) {
				break;
			}

			dataBufferLen_ += bytesRead;

			unsigned int offset = 0;
			while (offset + tsd.staticLength() < dataBufferLen_) {
				unsigned int dynamicLength = 0;

				size_t tupleLength = tsd.tupleLength(&dataBuffer_[offset]); // this->tupleLength(offset);
				if (offset + tupleLength <= dataBufferLen_) {
					offsets_.push_back(offset);
					offset += tupleLength;
				} else {
					break;
				}
			}

			std::sort(offsets_.begin(), offsets_.end(), TupleComparator(tsd, dataBuffer_));

			std::stringstream chunkFileName;
			chunkFileName << output_ << "." << chunkFileNames.size();
			chunkFileNames.push_back(chunkFileName.str());
			std::ofstream chunk(chunkFileName.str().c_str(), std::ios_base::out | std::ios_base::binary);
			tsd.writeHeader(chunk);

			std::vector<unsigned int>::const_iterator it = offsets_.begin(), end = offsets_.end();
			for (; it != end; ++it) {
				chunk.write(&dataBuffer_[*it], tsd.tupleLength(&dataBuffer_[*it]));
			}

			std::copy(dataBuffer_.begin() + offset, dataBuffer_.begin() + dataBufferLen_, dataBuffer_.begin());
			dataBufferLen_ -= offset;
			offsets_.clear();
		}
//		}

		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.0");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.1");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.2");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.3");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.4");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.5");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.6");
		//chunkFileNames.push_back("e:/wiki/ts/links_by_src.7");
		merge(tsd, chunkFileNames);
	}
private:
	typedef std::vector<std::string> StrList;

	class Chunk {
	public:
		const std::string& name;

		Chunk(TupleStreamDescription& tsd, const std::string& fileName): tsd_(tsd), stream_(fileName.c_str(), std::ios_base::in | std::ios_base::binary), buffer_(CHUNK_BUFFER_SIZE), bufferLen_(0), bufferPos_(0), name(fileName) { 
			TupleStreamDescription tmp(stream_, IndexList()); // eat the tuple stream header
		}

		char* currentTuple() {
			return &buffer_[bufferPos_];
		}

		bool hasMore() {
			if (remaining() < tsd_.staticLength()) {
				if (!fillBuffer()) {
					assert(remaining() == 0);
					return false;
				}
			}

			tupleLength_ = tsd_.tupleLength(&buffer_[bufferPos_]);
			if (remaining() < tupleLength_) {
				fillBuffer();
				assert(remaining() >= tupleLength_);
			}

			assert(remaining() == 0 || remaining() >= tupleLength_);

			return remaining() >= tupleLength_;
		}

		void write(std::ofstream& stream) {
			stream.write(&buffer_[bufferPos_], tupleLength_);
			bufferPos_ += tupleLength_;
		}

		std::string value() {
			return std::string(&buffer_[bufferPos_], tupleLength_);
		}
	private:
		bool fillBuffer() {
			if (bufferPos_ > 0) {
				std::copy(buffer_.begin() + bufferPos_, buffer_.begin() + bufferLen_, buffer_.begin());
				bufferLen_ -= bufferPos_;
				bufferPos_ = 0;
			}

			stream_.read(&buffer_[bufferLen_], buffer_.size() - bufferLen_);
			size_t bytesRead = stream_.gcount();

			if (bytesRead > 0) {
				bufferLen_ += bytesRead;
			}

			return bytesRead > 0;
		}

		size_t remaining() {
			return bufferLen_ - bufferPos_;
		}

		std::vector<char> buffer_;
		size_t bufferLen_, bufferPos_;
		std::ifstream stream_;
		TupleStreamDescription& tsd_;
		unsigned int tupleLength_;
	};

	struct MergeComparator {
		MergeComparator(TupleStreamDescription& tsd): tsd_(tsd) { }

		bool operator()(boost::shared_ptr<Chunk> x, boost::shared_ptr<Chunk> y) {
			int c = tsd_.compareTuples(x->currentTuple(), y->currentTuple());
//			tsd_.print(x->currentTuple());
//			tsd_.print(y->currentTuple());
//			std::cout << c << std::endl;
			return c > 0;
		}
	private:
		TupleStreamDescription& tsd_;
	};

	void merge(TupleStreamDescription& tsd, const std::vector<std::string>& chunkFileNames) {
		std::vector<boost::shared_ptr<Chunk>> heap;

		StrList::const_iterator it = chunkFileNames.begin(), end = chunkFileNames.end();
		for (; it != end; ++it) {
			boost::shared_ptr<Chunk> chunk(new Chunk(tsd, *it));
			if (chunk->hasMore()) {
				heap.push_back(chunk);
			}
		}

		MergeComparator comparator(tsd);

		make_heap(heap.begin(), heap.end(), comparator);

		std::ofstream outStream(output_.c_str(), std::ios_base::out | std::ios_base::binary);
		tsd.writeHeader(outStream);

		std::string v;

		while (!heap.empty()) {
			boost::shared_ptr<Chunk> chunk = heap.front();
			pop_heap(heap.begin(), heap.end(), comparator);

//			tsd.print(chunk->currentTuple());
//			tsd.print((*(heap.begin() + 1))->currentTuple());
//			std::cout << "---" << std::endl;

//			std::string& sv = chunk->value();
//			if (tsd.compareTuples(v.c_str(), sv.c_str()) > 0) {
//				std::cout << heap.size() << std::endl;
//				std::cout << v << " > " << sv << std::endl;
//			}
//			v = sv;

//			std::cout << chunk->name << std::endl;

			chunk->write(outStream);

			if (chunk->hasMore()) {
				push_heap(heap.begin(), heap.end(), comparator);
			} else {
				heap.erase(heap.end() - 1);
			}
		}
	}

	template<typename T> inline T readWord(char* data) {
		return *reinterpret_cast<T*>(data);
	}

	inline unsigned int fieldAsUInt(char* data, unsigned int fieldNum) {
		return readWord<unsigned int>(data + fieldNum * 4);
	}

	//	inline size_t tupleLength(unsigned int offset) {
	//	return tupleLength(&dataBuffer_[offset]);
	//}

	//size_t tupleLength(char* data) {
	//	size_t length = staticLength_;
	//	for (size_t i = 0; i < stringFields_; ++i) {
	//		length += fieldAsUInt(data, i);
	//	}
	//	return length;
	//}

	//const char* stringFieldPtr(char* data, size_t fieldNum) {
	//	size_t dynamicLength = stringFields_ * 4;
	//	int str_index = 0;
	//	for (size_t i = 0; i < fieldNum; ++i) {
	//		switch (fields_[i]) {
	//		case INTEGER:
	//			dynamicLength += 4;
	//			break;
	//		case STRING:
	//			dynamicLength += fieldAsUInt(data, str_index++);
	//			break;
	//		default:
	//			assert(false);
	//		}
	//	}
	//	return data + dynamicLength;
	//}

	//inline int compare(size_t offset1, size_t offset2) {
	//	return compare(&dataBuffer_[offset1], &dataBuffer_[offset2]);
	//}

	//int compare(char* x, char* y) {
	//	std::vector<unsigned int>::const_iterator it = comparatorFields_.begin(), end = comparatorFields_.end();
	//	for (; it != end; ++it) {
	//		size_t fieldNum = *it;

	//		unsigned int value1 = fieldAsUInt(x, fieldNum);
	//		unsigned int value2 = fieldAsUInt(y, fieldNum);

	//		long long c;
	//		switch (fields_[fieldNum]) {
	//		case INTEGER:
	//			c = value1 - value2;
	//			break;
	//		case STRING:
	//			{
	//				const char* str1 = stringFieldPtr(x, fieldNum),* str2 = stringFieldPtr(y, fieldNum);
	//				c = strncmp(str1, str2, std::min(value1, value2));
	//				if (c == 0) {
	//					c = value1 - value2;
	//				}
	//				//					std::cout << std::string(str1, value1) << "|||" << std::string(str2, value2) << std::endl;
	//			}
	//			break;
	//		default:
	//			assert(false);
	//		}

	//		if (c != 0) {
	//			return c < 0 ? -1 : 1;
	//		}
	//	}

	//	return 0;
	//}

	//inline unsigned int staticLength() {
	//	return staticLength_;
	//}

	IndexList comparatorFields_;
	std::vector<char> dataBuffer_;
	size_t dataBufferLen_;
	std::ifstream inStream_;
	const std::string input_, output_;
	std::vector<unsigned int> offsets_;
	//	unsigned int staticLength_;

	//	std::vector<FieldType> fields_;
	//	std::vector<unsigned int> comparatorFields_;
	//	size_t stringFields_;
};

//int _tmain(int argc, _TCHAR* argv[])
int main(int argc, char** argv) {
	try {
		if (argc < 3) {
			std::cout << "Usage: Sorter.exe [input file] [output file] [comparator fields]" << std::endl;
			return 1;
		}

		std::string input(argv[1]), output(argv[2]);

		IndexList comparatorFields;
		for (int i = 3; i < argc; ++i) {
			comparatorFields.push_back(atoi(argv[i]));
		}

//		TupleStreamSorter tss("e:/wiki/ts/links", "e:/wiki/ts/links_by_src");
//		TupleStreamSorter tss("c:/temp/ts/int_string_stream", "c:/temp/ts/int_string_stream_by_src");
		TupleStreamSorter tss(input, output, comparatorFields);
		tss.run();
	} catch (std::exception& e) {
		std::cout << "Terminal error: " << e.what() << std::endl;
	}

	int i;
	std::cin >> i;

	return 0;
}
