require 'WikiMiner/config'
require 'pp'

class EntityClassifier
  def initialize
    @stop_words = {}
    File.open('c:/Users/Christian/Documents/NetBeansProjects/WikiDump/lib/stop_words.txt', 'r') do |file|
      file.readlines.each do |stop_word|
        stop_word.strip!
        @stop_words[stop_word] = 1 if stop_word
      end
    end

    @min_capitalized_tokens = 2
    @min_double_capitalized_tokens = 2
    @min_exact_matches = 0.75
  end
  
  def candidate=(value)
    if /^([^\(\)]+)(?:\(([^\)]+)\))?$/ =~ value
      @candidate = $1.strip
      @domain = $2.try(:strip)
    else
      @domain = nil
      @candidate = value
    end
  end
  
  def text=(value)
    @text = value
  end
  
  def candidate; @candidate end
  def domain; @domain end
  def text; @text end
  
  def min_capitalized_tokens=(value); @min_capitalized_tokens = value end
  def min_double_capitalized_tokens=(value); @min_double_capitalized_tokens = value end
  def min_exact_matches=(value); @min_exact_matches = value end
  
  def exact_matches=(block)
    @exact_matches = block
  end
  
  def capitalized_in_text?
    return 0 if @text.nil?
    
    begin
      exact = 0
      nonexact = 0
    
      text = @text.gsub(/\[\[[^\]]*\]\]|\{[^\{\}]*\}/, '')
    
      return 0 if @candidate =~ /[\+\*\?\(\)\[\]\\]/ 
    
      text.scan(/(?:^|[^\w])(#{@candidate})(?:$|[^\w])/i) do |match|
        if @candidate == $1
          exact += 1
        else
          nonexact += 1
        end
      end
    
      return exact.to_f / (exact + nonexact).to_f
    rescue Exception => e
      puts "Bad candidate: #{candidate}\n#{e}"
      return 0
    end
  end
  
  def capitalized_tokens(tokens)
    #    tokens.inject(0) { |count, token| count + (token =~ /^[A-Z]/ ? 1 : 0) }
    tokens.inject(0) { |count, token| count + (token =~ /[A-Z]/ ? 1 : 0) }
  end

  def double_capitalized_tokens(tokens)
    tokens.inject(0) { |count, token| count + (token =~ /[A-Z].*[A-Z]/ ? 1 : 0) }
  end
  
  def classify
    tokens = @candidate.split(/ |-/)
    
    strip_stop_words(tokens)
    
    
    ct = capitalized_tokens(tokens)    
    return 1 if ct >= @min_capitalized_tokens and ct == tokens.length # >= @min_capitalized_tokens) ||
    return 2 if double_capitalized_tokens(tokens) >= @min_double_capitalized_tokens
    return 3 if @min_exact_matches > 0 && capitalized_in_text? >= @min_exact_matches
    #    return 3 if @exact_matches and @exact_matches.call(@candidate) >= @min_exact_matches

  end
  
  def strip_stop_words(tokens)
    tokens.collect! do |token|
      token unless @stop_words.has_key?(token) or !(token =~ /^[a-zA-Z]/)
    end
    tokens.compact!
  end
  
  def is_entity
    c = classify
    return !c.nil?
  end
end

if $0 == __FILE__
  ec = EntityClassifier.new
  tokens = ['The', 'Demonstration', 'of', 'Stop', 'Words']
  ec.strip_stop_words(tokens)
  pp tokens
  
  p ec.capitalized_tokens(tokens)
  p ec.double_capitalized_tokens(['the', 'OF', 'the', 'BesTest'])
  
  ec.exact_matches = Proc.new do |tokens|
    tokens == tokens.collect { |t| t.upcase } ? 1 : 0
  end
  
  ec.candidate = 'George Bush'
  pp ec.is_entity
  
  ec.candidate = 'Bush administration'
  pp ec.is_entity
  
  ec.candidate = 'Mercury (planet)'
  pp ec.candidate
  pp ec.domain
  
  puts "\n\n"
  ec.exact_matches = Proc.new { |tokens| tokens.index(' ') ? 0 : 1 }
  ec.candidate = 'Akers mekaniske verksted'
  pp ec.is_entity
  puts "\n\n"

  ec.candidate = 'Bush'
  pp ec.is_entity
  
  ec.candidate = 'List of assets owned by Time Warner'
  tokens = 'List of assets owned by Time Warner'.split(' ')
  ec.strip_stop_words(tokens)
  puts (tokens.is_a?(Array) and tokens.length > 1 ? 0 : 1)
  pp tokens
  pp ec.is_entity
end
