require 'tuple_stream'
require 'EntityCollector/entities'
require 'EntityCollector/synonyms'
require 'WikiMiner/tasks'
require 'ferret'
require 'EntityCollector/entity_classifier'
require 'Eval/SynEval'

class FindSynonyms
  def run
    @entities = {}

    Entities.open :mode => :read_all, :name => :from_categories do |name, full_name, domain, category, is_entity|
      @entities[full_name] = 1
    end
    
    Synonyms.open :mode => :write, :order_by => :synonym do |output|
      Synonyms.open :mode => :read_all, :order_by => :default do |synonym, main_synonym, type, weight|
        if @entities.has_key?(main_synonym)
          output.write(synonym, main_synonym, type, weight)
          #puts "#{synonym} - #{main_synonym} - #{weight} (#{type})"
        end
      end
    end
    #    Links.open :mode => :read_all do |src_namespace, src, dst_namespace, dst, caption, type, weight|
    #      
    #    end
  end
  
  def print_ordered(list)
    list.sort { |x, y| y[1] <=> x[1] }.each do |synonym, weight|
      puts "   #{synonym} - #{weight}"
    end
  end
  
  def split_array(array, &block)
    true_list = []
    false_list = []
    array.each do |element|
      if block.call(element)
        true_list << element
      else
        false_list << element
      end
    end
    return [true_list, false_list]
  end

  def print
    ec = EntityClassifier.new
    
    entities = {}
    Synonyms.open :mode => :read_all, :order_by => :synonym do |synonym, main_synonym, type, weight|
      entities[main_synonym] ||= {}
      entities[main_synonym][synonym] = weight
    end

    ordered = entities.sort { |x, y| y[1].length <=> x[1].length }
    (0..99).each do |i|
      total_weight = 0
      ordered[i][1].each { |key, value| total_weight += value }
      
      (a1, a2) = split_array(ordered[i][1]) do |synonym, weight|
        ec.candidate = synonym
        #        ec.min_capitalized_tokens = 
        ec.is_entity        
      end

      puts "#{ordered[i][0]} - #{ordered[i][1].length}"
      print_ordered a1
      puts "\n"
      print_ordered a2
      #      print_ordered ordered[i][1]
      #      ordered[i][1].sort { |x, y| y[1] <=> x[1] }.each do |synonym, weight|
      #        puts "   #{synonym} - #{weight}"# if weight > total_weight * 0.03
      #      end
      puts "\n\n"
    end
  end

  def index
    puts "Reading entities...\n"
    entities = {}
    Synonyms.open :mode => :read_all, :order_by => :synonym do |synonym, main_synonym, type, weight|
      entities[main_synonym] ||= {}
      entities[main_synonym][synonym] = weight
    end

    puts "Indexing...\n"

    Ferret::Index::Index.new(:path => 'e:/wiki/index', :analyzer => Ferret::Analysis::WhiteSpaceAnalyzer.new) do |index|
      entities.each do |main_synonym, synonyms|
        total_weight = 0
        synonyms.each { |_, value| total_weight += value }

        synonyms = synonyms.collect do |synonym, weight|
          #          if weight > 5 or weight > total_weight * 0.01
          synonym = main_synonym if synonym == ''
          "SYNSYN #{synonym.downcase} SYNSYN #{weight}"
          #          end
        end.compact.join("\n")
        index << {:main_synonym => main_synonym, :synonyms => synonyms, :total_weight => total_weight}
      end
    end
  end
  
  def find_synonyms(classifier, s)
    # calculate weights
    weights = Hash.new(0)
    s[:links].each { |synonym, weight| weights[synonym.downcase] += weight }
    s[:redirects].each { |synonym| weights[synonym.downcase] += 1 }

    total_weight = weights.inject(0) { |w, syn| w + syn[1] }
    min_weight = total_weight / 20.0

    synonyms = {}
    s[:links].each do |synonym, weight|
      classifier.candidate = synonym
      if classifier.is_entity or weight > min_weight
        sd = synonym.downcase
        synonyms[sd] = weights[sd]
      end
    end

    s[:redirects].each do |synonym|
      classifier.candidate = synonym
      classifier.min_capitalized_tokens = 1
      sd = synonym.downcase
      if classifier.is_entity or weights[sd] > 1
        synonyms[sd] = weights[sd]
      end
    end

    if weights.has_key?('')
      synonyms.delete('')
      synonyms[s[:name]] = weights['']
    end
    
    return synonyms
  end
  
  def index2
    puts "Reading entities...\n"
    entities = SynEval.get_entities { |name, category| category == 'person' || category == 'organization' || category == 'company' }
    
    puts "Indexing...\n"
    
    classifier = EntityClassifier.new
    
    field_infos = Ferret::Index::FieldInfos.new(:term_vector => :no)
    field_infos.add_field(:name)
    field_infos.add_field(:spec, :index => :no, :store => :compressed)
    field_infos.add_field(:synonyms, :term_vector => :with_offsets)
    field_infos.add_field(:weight, :index => :untokenized)
    
    Ferret::Index::Index.new(
      :analyzer => Ferret::Analysis::WhiteSpaceAnalyzer.new,
      :create => true,
      :path => 'e:/wiki/index',
      :field_infos => field_infos,
      :max_buffer_memory => 0x4000000,
      :max_buffered_docs => 50000,
      :merge_factor => 20
    ) do |index|
      SynEval.read_synonyms() do |s|
        if entities.has_key?(s[:name])
          spec = PP.pp(s, "")

          synonyms = find_synonyms(classifier, s)
          
          if synonyms.length >= 5
            weight = synonyms.inject(0) { |w, syn| w + syn[1] }
            
            ##            pp weights
            ##            pp synonyms

            #          weight = s[:links].inject(0) { |w, syn| w += syn[1] }

            synonym_text = synonyms.sort { |x, y| y[1] <=> x[1] }.
              collect { |synonym, w| "SYN #{synonym} SYN #{w}" }.join("\n")

            index << {:category => entities[s[:name]], :name => s[:name], :spec => spec, :synonyms => synonym_text, :weight => weight}
            
            puts "'#{s[:name]}' (#{synonyms.length}) (w=#{weight}) indexed."
          end
        end
      end
    end
  end
end

if $0 == __FILE__
  fs = FindSynonyms.new
  #  fs.run
  #fs.print
  #  fs.index
  fs.index2
end
