require 'WikiMiner/db'
require 'tuple_stream'
require 'WikiMiner/tasks'

class Synonyms < TupleStream
  define_field :synonym, :string
  define_field :main_synonym, :string
  define_field :type, :string
  define_field :weight, :integer

  define_comparator :default, [:main_synonym, :type, :synonym]
  define_comparator :synonym, [:synonym, :main_synonym]
end

class SynonymsCollector
  def flush(output, main_synonym, weights)
    unless main_synonym.nil?
      weights.each do |id, weight|
        synonym, type = id
        #        puts "#{synonym}|||#{type}|||#{weight}\n"
        output.write(synonym, main_synonym, type, weight)
      end
    end
  end

  def run
    #    if 1 == 0
    Synonyms.open :mode => :write, :name => :ungrouped_synonyms, :order_by => :default do |output|
      Links.open :mode => :read_all, :order_by => :dst do |src_namespace, src, dst_namespace, dst, caption, type, weight|
        if dst_namespace == 0
          output.write(type == 'link' ? caption.to_s : src.to_s, dst.to_s, type.to_s, weight.to_i) if dst_namespace == 0
        end
      end
    end
    #    end

    Synonyms.open :mode => :write, :order_by => :default do |output|
      ms = nil
      w = Hash.new(0)
      Synonyms.open :mode => :read_all, :name => :ungrouped_synonyms, :order_by => :default do |synonym, main_synonym, type, weight|
        if ms != main_synonym
          flush(output, ms, w)
          ms = main_synonym
          w = Hash.new(0)
        end

        synonym = '' if type == 'link' and main_synonym.casecmp(synonym) == 0
      
        synonym.gsub!(/\([a-zA-Z0-9 _]+\)$/, '')
        synonym.strip!
        synonym.gsub!(/['´`]s$/, '')
        synonym.gsub!(/["']/, '')
        synonym.strip!
        w[[synonym, type]] += weight
      end
      flush(output, ms, w)
    end
  end
end

if $0 == __FILE__
  s = SynonymsCollector.new
  s.run
end
