# 
# To change this template, choose Tools | Templates
# and open the template in the editor.

require 'singleton'

module Wiki
  def cleanup_name!(name)
    name.gsub!(/\#.*/, '')
    name.gsub!(/_/, ' ')
    name.strip!
    name[0] = name[0,1].upcase
  end

  def cleanup_name(name)
    n = name.clone
    cleanup_name!(name)
  end
  
  module_function :cleanup_name!, :cleanup_name
end

class DatabaseLink
  def initialize
    up unless up?
  end
  
  def db
    @db ||= Config[:db]
  end
  
  def up?
    false
  end
  
  def down?
    !up?
  end
  
  def table_exists?(table_name)
    db.get_first_value("SELECT name FROM sqlite_master WHERE name = :table_name", 
      'table_name' => table_name) == table_name
  end
  
  def drop(*args)
    args.each { |name| db.execute("DROP TABLE IF EXISTS #{name}") }
  end
  
  def sql(statement, *args)
    db.execute(statement, args)
  end
  
  def reset; down; up end
end

class Redirects < DatabaseLink
  def initialize
    super
  end
  
  def finish
    @stmt.close if @stmt
  end
  
  def add_redirect(source, target)
    Wiki::cleanup_name!(source)
    Wiki::cleanup_name!(target)
    
    @stmt = db.prepare('INSERT INTO Redirect(Source, Target) VALUES(:Source, :Target)') unless @stmt
    @stmt.execute(:Source => source, :Target => target) 
  end
  
  def parse_redirect(source, text)
    if text =~ /^\#REDIRECT.*?\[\[([^\]\|]+)(?:|[^\]]*)\]\]/i
      add_redirect(source, $1)
    end
  end
  
  def up?
    table_exists?('Redirect')
  end
  
  def down; drop('Redirect') end
  def up;
    sql('CREATE TABLE IF NOT EXISTS Redirect(Source VARCHAR(127), Target VARCHAR(127), PRIMARY KEY(Source, Target))')
  end
end

class Templates < DatabaseLink
  def initialize
    super
  end
  
  def finish
    @stmt.close if @stmt
  end
  
  def add_template(name, text)
    Wiki::cleanup_name!(name)
    puts name
    @stmt = db.prepare('INSERT INTO Template(TemplateName, TemplateText) VALUES (:Name, :Text)') unless @stmt
    @stmt.execute(:Name => name, :Text => text)
  end
  
  def extract_template(name, text)
    if name =~ /^Template:(.+)/
      add_template($1, text)
    end
  end
  
  def up?; table_exists?('Template') end
  def down; drop('Template') end
  def up; sql('CREATE TABLE IF NOT EXISTS Template(TemplateName VARCHAR(127), TemplateText TEXT, PRIMARY KEY(TemplateName))') end
end

#module Singleton
#  def self.global
#    @instance ||= new
#  end
#end

class TemplateCache < DatabaseLink
  include Singleton
  
  def initialize
    @cache = {}
    #    @cache = Hash.new do |hash, key|
    #      @stmt = db.prepare('SELECT Template FROM Template WHERE Name = :Name') unless @stmt
    #      row = @stmt.execute(:Name => key)
    #      hash[key] = row['Template']
    #    end
  end
  
  def [](key)
    puts "#{key}"
    @cache.fetch(Wiki::cleanup_name(key)) do |key|
      @stmt = db.prepare('SELECT Template FROM Template WHERE Name = :Name') unless @stmt
      row = @stmt.execute!(:Name => key)
      @cache[key] = row.length > 0 ? row[0]['Template'] : nil
    end
  end
end

class PageRenderer
  def self.render(title, text)
    text.gsub(/\{\{([\w\s]+)[^\}]*\}\}/) { TemplateCache.instance[$1] }
  end
end
