require 'wiki_page_reader'
require 'WikiMiner/binary_dump'
require 'WikiMiner/config'
require 'WikiMiner/names'
require 'WikiMiner/task_manager'
require 'WikiMiner/redirects'
require 'WikiMiner/tasks'
require 'WikiMiner/wikipedia'
require 'oniguruma'
require 'tuple_stream'

TaskManager.default :only => [:write_titles, :write_shuffled_titles, :write_category_pages, :write_links_and_redirects, :write_entities]
TaskManager.register do
  bind :names => Names.new
  bind :pages => WikiPageReader.new("#{Config[:root]}all_pages.dat")
  bind :parse_redirects => Redirects.new
  bind :extract_templates => Templates.new
  
  #  bind :write_binary_dump => DumpWriter.
  
  stage(:read_xml,
    task(:write_binary_dump)
  )
  
  stage(:read_binary_dump,
    task(:write_templates),
    task(:write_topics)
  )
  
  stage(:read_binary_dump,
    task(:write_links_and_redirects),
    task(:write_categories),
    task(:write_entities),
    task(:write_titles),
    task(:write_category_pages)
  )
  
  stage(:null_source,
    task(:write_shuffled_titles)
  )

  stage(:pages,
    task(:names) { |t, page| t << page[:title] },
    task(:dummy) do |t, page|
      title = page[:title]
      if title =~ /^[\w+.\s]+$/
        page = page[:text]
        page.gsub!(/\[[^\[\]]?\]/, '')
            
        #        r = Oniguruma::ORegexp.new("((?=[\w\s]{0,5}))(#{title})")
        r = Oniguruma::ORegexp.new("(?i)(#{title})")
            
        entity = non_entity = possible_entity = 0        
        #        page.scan(/([\w\s]{1,5})?(#{title})/i) do |match|
            
            
        r.scan(page) do |match|
          if match[0] == title
            if /(\w|.)\s+$/.match(page[0, match.offset(0)[0]])
              if $1
                entity += 1
              else
                possible_entity += 1
              end
            else
              possible_entity += 1
            end
            #            if match[0]
            #              entity += 1
            #            else
            #              possible_entity += 1
            #            end
          else
            non_entity += 1
          end
          #          puts match.offset(0).join(',')
          #          puts "#{match[0]}@@@#{match[1]}\n"
          #          puts match.join('@@@')
        end
            
        puts "#{title}: #{entity}/#{possible_entity}/#{non_entity}"
      else
        puts "Ignored: #{title}"
      end
    end,
    task(:parse_redirects) { |t, page| t.parse_redirect(page[:title], page[:text]) },
    task(:extract_templates) { |t, page| t.extract_template(page[:title], page[:text]) },
    task(:render_pages) { |t, page| PageRenderer.render(page[:title], page[:text]) }
  )
end
TaskManager.run
