require 'pqueue'

class UnknownKeyException < Exception
  
end

class TupleBuffer
  def initialize(max_size = 64 * 1024 * 1024)
    @max_size = max_size
  end
  
  def self.fill(reader, max_size = 64 * 1024 * 1024)
    buffer = []
    len = 0
    while item = reader.read and len < max_size
      item.each_pair { |key, value| len += value.respond_to?(:length) ? value.length + 50 : 60 }
      buffer << item
    end
    return buffer
  end
end

class BinaryStream
  @@comparators = {}
  @@fields = {}
  
  def initialize(fields, io)
    @fields = fields
    @io = io
    init
  end
    
  def self.field(*args)
    type, name, *options = args
    
    @@fields[self.name] ||= {}
    @@fields[self.name][name] = type
  end
  
  def self.comparator(name, fields = [], &block)
    c = @@comparators[self.name] ||= {}
    if block_given?
      c[name] = block
    else
      puts "lambda { |x, y| #{fields.collect { |f| "(x[:#{f.to_s}] <=> y[:#{f.to_s}])" }.join(' || ')}}\n"
      comp = "{ |x, y| #{fields.collect { |f| "(x[:#{f.to_s}] <=> y[:#{f.to_s}]).nonzero?" }.join(' || ')}}"
      c[name] = [eval("lambda #{comp}"), eval("lambda { |arr| arr.sort! #{comp} }")]
    end
  end

  def self.open(options = {}, &block)
    #    io = nil
    #    io_name = nil
    #    
    #    args.each do |option|
    #      key, value = option.to_a.flatten
    #      case key
    #      when :source
    #        raise "source/target already set" if io
    #        io_name = value
    #        io = File.new("c:/temp/bs/#{value}", 'rb')
    #      when :target
    #        raise "source/target already set" if io
    #        io_name = value
    #        io = File.new("c:/temp/bs/#{value}.tmp", 'wb')
    #      end
    #    end

    raise "Only a single source or target must be specified" if options[:source] and options[:target]

    name = "c:/temp/bs/"
    if options[:source]
      name += options[:source]
      mode = 'rb'
    elsif options[:target]
      name += "#{options[:target]}.tmp"
      mode = 'wb'
    else
      raise "Missing source or target"
    end
    
    io = File.new(name, mode)
    
    obj = new(@@fields[self.name].to_a, io)
    
    if block_given?
      begin
        yield obj
      ensure
        io.close
      end
      File.rename(name, "c:/temp/bs/#{options[:target]}") if options[:target]
    else
      obj
    end
  end
  
  def self.sorter(options = {})
    #    options = args.collect { |i| i.to_a }.
    #    batch_size = options[:batch_size] or 10000
    comparator, sorter = @@comparators[self.name][options[:comparator] || :default]
    
    #    puts comparator.call({:src => 55, :dst => 54, :caption => ''}, {:src => 55, :dst => 54, :caption => ''})
    #    exit
    
    open :source => options[:source] do |input|
      temp_files = 0
      
      timer :sort_files do
        while !input.eos?
          buffer = nil
          timer :fill_buffer do
            buffer = TupleBuffer.fill(input)
          end
          
          timer :sort_buffer do
            sorter.call(buffer)
            #            buffer.sort! { |x, y| comparator.call(x, y) }
          end
        
          timer :flush_buffer do
            unless input.eos? and temp_files == 0 and !buffer.empty?
              open :target => options[:target] + ".#{temp_files}" do |temp|
                buffer.each { |item| temp << item }
              end
              temp_files += 1
            end
          end
          
          puts "\n"
        end
      end

      timer :merge do
        open :target => options[:target] do |output|
          if temp_files > 0
            files = (0..temp_files - 1).collect { |i| open :source => options[:target] + ".#{i}" }
         
            queue = PQueue.new lambda { |x, y| comparator.call(x[0], y[0]) < 0 }
            files.each { |f| queue.push([f.read, f]) }

            until queue.empty?
              item, file = queue.pop
              output << item            
              next_item = file.read
              queue.push([next_item, file]) if next_item
            end
          
          
            files.each { |f| f.close }
            #          (0..temp_files - 1).each { |i| File.delete("c:/temp/bs/#{options[:target]}.#{i}") }
          else
            buffer.each { |item| output << item }
          end
        end
      end
    end
  end
  
  #  def self.open(name, mode)
  #    File.open("c:/temp/bs/#{name}", mode) do |f|
  #      bs = new(@@fields[self.name].to_a, f)
  #      yield bs
  #    end
  #  end
  #  
  #  def self.input(name, &block)
  #    open name, 'rb', &block
  #  end
  #  
  #  def self.output(name, &block)
  #    open name, 'wb', &block
  #  end
  
  def append_int(buffer, value)
    buffer << [value].pack('V')
  end
  
  def close
    @io.close
  end
  
  def init
    write = <<str
    def <<(item)
      buffer = []
      len = 0
str
    @fields.each do |key, type|
      case type
      when :integer
        write += "append_int buffer, item[:#{key}]; len += 4;\n"
      when :string
        write += "str = item[:#{key}]; append_int buffer, str.length; buffer << str; len += 4 + str.length;\n"
      end
    end
    write += <<str
    @io.write([len].pack('V'))
    @io.write(buffer.join)
  end
str
    eval(write)
  end
  
  #  def <<(item)
  #    buffer = []
  #    len = 0
  #    
  #    #    item.each_pair do |key, value|
  #    @fields.each do |key, type|
  #      value = item[key]
  #      
  #      case type
  #      when :integer
  #        append_int buffer, value
  #        len += 4
  #      when :string
  #        append_int buffer, value.length
  #        buffer << value
  #        len += 4 + value.length
  #      end
  #    end
  #
  #    @io.write([len].pack('V'))
  #    @io.write(buffer.join)
  #  end
  
  def >>(item)
    len = @io.read(4).unpack('V').first
    offset = 0
    packed_item = @io.read(len)
    
    @fields.each do |key, type|
      case type
      when :integer
        item[key] = packed_item[offset..offset + 3].unpack('V').first
        offset += 4
      when :string
        len = packed_item[offset..offset + 4].unpack('V').first
        #        puts "len: #{len}\n"
        item[key] = packed_item[offset + 4..offset + 4 + len - 1]
        offset += 4 + len
      end
    end
  end
  
  def read
    return nil if @io.eof?
    item = {}
    self.>> item
    return item
  end
  
  def eos?; @io.eof? end

  private :initialize
end

class Category < BinaryStream
  field :string, :parent
  field :string, :child
  field :string, :caption
end

class Template < BinaryStream
  field :string, :name
  field :string, :template
end

class Links < BinaryStream
  field :integer, :src
  field :integer, :dst
  field :string, :caption
  
  comparator :default, [:src, :dst, :caption]
  comparator :secondary do |x, y| x[:src] <=> y[:src] or x[:dst] <=> y[:dst] or x[:caption] <=> y[:caption] end
end

def timer(name)
  start = Time.now
  yield
  stop = Time.now
  puts "#{name} in #{stop - start} secs"
end

p ["test"].pack('a')

t = Time.now

timer :write do
  Links.open :target => 'links' do |l|
    5000000.times do |i|
      l << {:src => i % 42 , :dst => 1024 * (i % 12) + i, :caption => 'This is the caption!'}
    end
  end
end

timer :sort do
  Links.sorter :source => 'links', :target => 'links2', :comparator => :default
end

exit

Links.open :source => 'links2' do |l|
  10000.times do |i|
    item = l.read
    p item if item
  end
end

exit

Category.output 'test' do |c|
  100000.times { c << {:parent => 'parent category', :child => 'child category', :caption => 'capt'} }
end

puts "Took #{Time.now - t} secs"

Category.input 'test' do |c|
  item = {}
  100000.times { c >> item }
  #  p item
end

puts "Took #{Time.now - t} secs"

#Template.open << {:name => 'gerijretert'}
