require 'wiki_page_reader'
require 'pp'
require 'sqlite3'

def db
  if !$conn
    $conn = SQLite3::Database.new('E:\\wiki\\wiki.db', {})
    $conn.busy_timeout(15000)
    $conn.results_as_hash = true
    
    #    $conn.execute_batch <<SQL 
    #          DROP TABLE IF EXISTS Pages;
    #          CREATE TABLE IF NOT EXISTS Pages(PageId INT, PageTitle 
    #          VARCHAR(127), PageType VARCHAR(1), PRIMARY KEY(PageId));
    #    
    #          DROP TABLE IF EXISTS SubCategories;
    #          CREATE TABLE IF NOT EXISTS SubCategories(ParentId INT, ChildId INT, PRIMARY KEY(ParentId, ChildId));
    #    
    #          DROP TABLE IF EXISTS Categories;
    #          CREATE TABLE IF NOT EXISTS Categories(CategoryId INT NOT NULL, CategoryLevel INT DEFAULT NULL, PRIMARY KEY(CategoryId));
    #SQL
  end
  $conn
end


def each_category(text)
  result = [] unless block_given?
  
  text.scan %r{\[\[(Category:[^|\]]+)(?:|[^\]]*)?\]\]} do |name|
    name[0].strip!
    if block_given?
      yield name[0]
    else
      result << name[0]
    end
  end
  
  result unless block_given?
end

def h; Hash.new { |hash, key| hash[key] = [] } end

def get_page_type(page_title)
  case page_title
  when /^Category:/
    'c'
  when /^Template:/
    't'
  else
    'p'
  end
end

def lookup_id(hash, key)
  value = hash[key]
  if !value
    hash[key] = hash.length
  else
    value
  end
end

def extract_categories
  $category_pages = h
  $titles = {}

  reader = WikiPageReader.new("e:/wiki/all_pages.dat")
  reader.each do |p|
    $child_id = lookup_id($titles, p[:title].strip)
    each_category(p[:text]).uniq.each do |category_name|
      $parent_id = lookup_id($titles, category_name)
      $category_pages[$parent_id] << $child_id
    end
  end
  
  db.transaction do |tran|
    tran.prepare("INSERT INTO Pages(PageId, PageTitle, PageType) VALUES(:page_id, :page_title, :page_type)") do |stmt|
      $titles.each do |page_title, page_id|
        stmt.execute :page_id => page_id, :page_title => page_title, :page_type => get_page_type(page_title)
      end
    end
  
    tran.prepare("INSERT INTO SubCategories(ParentId, ChildId) VALUES (:parent_id, :child_id)") do |stmt|
      $category_pages.each do |parent_id, children|
        children.each do |child_id|
          #          puts "#{parent_id} => #{child_id}"
          stmt.execute :parent_id => parent_id, :child_id => child_id
        end
      end
    end
  end
  
  #  category_pages.default = nil
  #  parent_categories = nil
  #  
  #  File.open('e:/wiki/category_pages.dat', 'wb') do |file|
  #    Marshal.dump(category_pages, file)
  #    Marshal.dump(titles, file)
  #  end
  #
  #  File.open('e:/wiki/category_graph.dat', 'wb') do |file|
  #    Marshal.dump(parent_categories, file)
  #  end
end

def Node
  def initialize(id); @id = id end
  
  def id; @id end
  def children; @children end
  def children=(children); @children = children end
  def parents; @parents end
  def parents=(parents); @parents = parents end
end

class Categories
  def read
    if !@category_pages || !@titles
      File.open('e:/wiki/category_pages.dat', 'rb') do |file| 
        @category_pages = Marshal.load(file)
        titles_lookup = Marshal.load(file)
        
        @titles = titles_lookup.sort { |x, y| x[1] <=> y[1] }.collect { |key, value| key}
      end
    end
  end
  
  def cp
    @category_pages || (read; @category_pages)
  end
  
  def t
    @titles || (read; @titles)
  end
  
  def find_leaves
    read
    
    puts "Number of categories: #{cp.length}"
    
    leaves = []
    
    cp.each do |parent_category, values|
      children = values.collect do |title_id|
        title = @titles[title_id]
        if title =~ /^Category:/
          title_id
        end
      end
      children.compact!
      leaves << parent_category if children.empty? 
    end
    
    puts "Leaf categories: #{leaves.length}"
    
  end
  
  def show_info
    cp.sort { |x, y| x[1].length <=> y[1].length }.each do |key, value|
      puts "#{key}: #{value.length}\n" if value.length > 50
    end
  end
end

def find_levels
  db.execute("DELETE FROM Categories")
  db.execute("INSERT INTO Categories(CategoryId) SELECT PageId FROM Pages WHERE PageType = 'c'")
  
  categories = db.execute("SELECT CategoryId, CategoryLevel FROM Categories")
  
  puts "Categories: #{categories.length}"
  
  level = 0
  done = false
  while !done
    #    db.execute("UPDATE Categories SET CategoryLevel = :level WHERE CategoryLevel IS NULL AND CategoryId NOT IN (SELECT ParentId FROM SubCategories WHERE ChildId IN (SELECT CategoryId FROM Categories WHERE CategoryLevel IS NULL))", 'level' => level)
    #    rows = db.changes

    $excluded = {}
    
    db.execute("SELECT DISTINCT(ParentId) AS ParentId FROM SubCategories WHERE ChildId IN (SELECT CategoryId FROM Categories WHERE CategoryLevel IS NULL)") do |row|
      $excluded[row['ParentId'].to_i] = 1
    end
    
    #    categories.each do |row|
    #      if $excluded[row['CategoryId'].to_i]
    #        puts "!!!\n"
    #      end
    #    end
    
    #    pp $excluded
    
    puts "Excluded: #{$excluded.length}"
    
    rows = 0
    
    db.transaction do |tran|
      tran.prepare("UPDATE Categories SET CategoryLevel = :level WHERE CategoryId = :id") do |stmt|
        categories.each do |row|
          #          pp row
          #          puts "#{$excluded[row['CategoryId']]}\n"
          if !row['CategoryLevel'] and !$excluded.has_key?(row['CategoryId'].to_i)
            stmt.execute 'id' => row['CategoryId'], 'level' => level
            row['CategoryLevel'] = level
            rows += 1
          end
        end
      end
    end
    
    done = rows == 0
    puts "Level #{level} categories: #{rows}\n"
    level += 1
  end  
end

def table_exists?(table_name)
  db.get_first_value("SELECT name FROM sqlite_master WHERE name = :table_name", 
    'table_name' => table_name) == table_name
end

def test
  begin
    if table_exists?("test")
      puts "exists"
    end
    
    db.execute("EXPLAIN SELECT COUNT(*) FROM test")
  rescue SQLite3::SQLException
    puts "Non-existing table"
  end
end

#extract_categories
#show_info
  
#c = Categories.new
#c.find_leaves

#find_levels

test
