class CreateDb < ActiveRecord::Migration
  def self.up
    create_table :tasks, :primary_key => :name do |t|
      t.string :name
      t.integer :status
    end
    add_index :tasks, :name, :unique
    
    create_table :templates do |t|
      t.string :name
      t.binary :template
    end
    add_index :templates, :name, :unique
    
    create_table :topics do |t|
      t.integer :namespace
      t.string :name
      t.column :flags, :integer, :default => 0
      t.column :has_content, :boolean, :default => false
    end
    add_index :topics, [:namespace, :name], :unique
    
    create_table :categories, :id => false, :primary_key => [:parent_id, :child_id] do |t|
      t.integer :parent_id, :child_id
    end
    add_index :categories, [:parent_id, :child_id], :unique
  end
  
  def self.down
    drop_table :tasks
    drop_table :templates
    drop_table :topics
    drop_table :categories
  end
end
