#require 'unprof'
require 'WikiMiner/config'
require 'WikiMiner/wikipedia'
require 'wiki_xml_parser'
require 'pqueue'

class Array
  def index_of(*value)
    if block_given?
      each_index { |i| return i if yield(self[i]) }
    else
      each_index { |i| return i if self[i] == value }
    end
  end
end

class TupleBuffer
  def initialize(max_size = 64 * 1024 * 1024)
    @max_size = max_size
  end
  
  def self.fill(reader, max_size = 384 * 1024 * 1024)
    buffer = []
    len = 0
    while item = reader.read
      item.each { |value| len += value.respond_to?(:length) ? value.length + 50 : 60 }
      buffer << item
      break if len >= max_size
      #      GC.start if buffer.length % 10000 == 0
    end
    return buffer
  end
end

def timer(name)
  start = Time.now
  result = yield
  stop = Time.now
  puts "#{name} in #{stop - start} secs"
  result
end

class TupleStream
  @@fields = {}
  @@comparators = {}
  @@options = {:root => 'e:/wiki/ts/', :sorter => 'e:/wiki/src/WikiMiner/Release/Sorter.exe'}
  
  def self.options; @@options end
  
  def init(name, action)
    raise Exception, "Unknown action" unless [:read, :read_all, :write, :write_all].include?(action)
    
    @name = name
    @action = action
    @fields = @@fields[self.class.name]
    @comparators = @@comparators[self.class.name]
    
    make_read_write
    open
  end
  
  def new_method(name, attrs, *args)
    attrs ||= ''
    self.instance_eval("def #{name.to_s}(#{attrs}); #{args.join(' ')} end")
  end
  
  def make_read_write
    strlens = @fields.collect { |name, type| "al_#{name}" if type == :string}.compact
    unpack = @fields.collect do |name, type|
      case type
      when :integer
        'V'
      when :string
        'A#{al_' + name.to_s + '}'
      end
    end
    
    new_method(:read, nil,
      strlens.empty? ? '' : "strlens = @io.read(#{strlens.length * 4}); return nil unless strlens;",
      strlens.empty? ? '' : "(#{strlens.join(',')},) = strlens.unpack('#{'V' * strlens.length}');",
      #      strlens.empty? ? "len = #{(@fields.length - strlens.length) * 4};" : "len = #{([(@fields.length - strlens.length) * 4] + strlens).join('+')};",
      "len = #{([(@fields.length - strlens.length) * 4] + strlens).join('+')};",
      "data = @io.read(len); return nil unless data; data.unpack(\"#{unpack}\");"
      #      "[#{strlens.join(',')}] = @io.read(#{strlens.length * 4})"
    )
    
    packarg = {:integer => 'V', :string => 'A*'}
    
    data = (@fields.collect { |name, type| "a_#{name.to_s}.length" if type == :string } + 
        @fields.collect { |name, type| "a_#{name.to_s}" }).compact.join(',')
    pack = (@fields.collect { |name, type| 'V' if type == :string } +
        @fields.collect { |name, type| packarg[type] }).compact.join
    
    new_method(:write, 
      @fields.collect { |name, _| "a_#{name.to_s}" }.join(', '),
      "@io.write([#{data}].pack('#{pack}'));"
    )

    data = ((0..@fields.length - 1).collect  { |index| "arr[#{index}].length" if @fields[index][1] == :string } +
        (0..@fields.length - 1).collect { |index| "arr[#{index}]" }).compact.join(',')
    pack = (@fields.collect { |name, type| 'V' if type == :string } +
        @fields.collect { |name, type| packarg[type] }).compact.join

    new_method(:write_array,
      'arr',
      "@io.write([#{data}].pack('#{pack}'));"
    )
  end
  
  def self.define_field(name, type)
    arr = @@fields[self.name] ||= []
    arr << [name, type]
  end
  
  def self.define_comparator(name, fields)
    hash = @@comparators[self.name] ||= {}
    fields = [fields] unless fields.is_a?(Array)
    hash[name] = fields
  end
  
  def get_comparator_code(name, var1 = 'x', var2 = 'y')
    @comparators[name].collect do |field| 
      index = get_field_index(field)
      "(#{var1}[#{index}] <=> #{var2}[#{index}])" + (index < @comparators[name].length - 1 ? ".nonzero?" : "")
    end.join(' || ')
  end
  
  def get_array_sorter(name)
    eval("lambda { |arr| arr.sort! { |x, y| #{get_comparator_code(name)} } }")
  end
  
  def get_item_comparator(name, var1 = 'x', var2 = 'y')
    eval("lambda { |x, y| (#{get_comparator_code(name, var1, var2)}) < 0 }")
  end
  
  def get_field_index(name)
    @fields.each_index { |i| return i if @fields[i][0] == name }
  end
  
  def get_chunk_name(num)
    "#{@name}_#{num}"
  end
  
  def self.sort(args = {})
    target_order_by = args[:target_order_by] || args[:order_by] || :default
    
    open :name => args[:source], :order_by => args[:source_order_by] do |input|
      open :name => args[:target], :mode => :write, :order_by => target_order_by do |output| 
        output.sort_from(input, target_order_by)
      end
    end
  end
  
  def self.sort2(args = {})
    target_order_by = args[:target_order_by] || args[:order_by] || :default
    
    comparators = @@comparators[self.name]
    comparator = comparators[target_order_by]
    fields = @@fields[self.name]
    
    input = full_name(args[:source], args[:source_order_by])
    output = full_name(args[:target], target_order_by)

    comparator_fields = comparator.collect { |x| fields.index_of { |y| p y; x == y[0] } }
    p comparator_fields
    
    system("\"#{options[:sorter]}\" \"#{input}\" \"#{output}\" #{comparator_fields.join(' ')}")
  end
  
  def sort_from(input, comparator)
    chunks = []
    sorter = get_array_sorter(comparator)

    begin
      until input.eos?
        buffer = timer(:fill) { TupleBuffer.fill(input) }
        timer(:sort) { sorter.call(buffer) }
          
        timer(:flush) do
          unless input.eos? and chunks.empty?
            chunk_name = get_chunk_name(chunks.length)
            self.class.open :full_name => chunk_name, :mode => :write do |chunk|
              buffer.each { |item| chunk.write_array(item) }
            end
            chunks << chunk_name
          end
        end
      end
      
      timer(:merge) do
        unless chunks.empty?
          queue = PQueue.new(get_item_comparator(comparator, 'x[0]', 'y[0]'))

          chunk_streams = chunks.collect { |name| self.class.open :full_name => name }
          begin
            chunk_streams.each { |stream| queue.push([stream.read, stream]) }
          
            until queue.empty?
              item, stream = queue.pop
              write_array(item)
              next_item = stream.read
              queue.push([next_item, stream]) if next_item
            end
          ensure
            chunk_streams.each { |stream| stream.close }
          end
        else
          buffer.each { |item| write_array(item) }
        end
      end
    ensure
      chunks.each { |chunk| File.delete("#{@@options[:root]}#{chunk}") }
    end
  end
  
  def self.full_name(name, order_by)
    full_name = self.name.gsub(/([A-Z])/, ' \1').strip.gsub(' ', '_').downcase
    full_name += '-' + name.to_s if name
    full_name += '-by-' + order_by.to_s if order_by
    return (@@options[:root] || '') + full_name
  end
  
  def self.open(args = {})
    action = args[:mode] || :read
    name = args[:full_name] || full_name(args[:name], args[:order_by])

    stream = self.new
    stream.init(name, action)

    return stream if [:read, :write].include?(action) and !block_given?

    begin
      case action
      when :read, :write
        yield stream
      when :read_all
        if block_given?
          until (item = stream.read).nil? 
            yield(item)
          end
        else
          items = []
          until (item = stream.read).nil?
            items << item
          end
          return items
        end
      when :write_all
        if block_given?
          until (item = yield).nil?
            stream.write_array(item)
          end
        end
      end
      
      stream.close
    ensure
      stream.close(false) unless stream.closed?
    end
  end

  def open
    @filename = @name
    
    case @action
    when :read, :read_all
      set_mode('rb')
      read_header
    when :write, :write_all
      @filename += '.wrk'
      set_mode('wb')
      write_header
    end
  end
  
  def set_mode(mode)
    @io = File.new(@filename, mode)
  end
  
  @@header = ['TUPLESTREAM', 34, 42, 111, 247].pack("A*CCCC")
  
  def create_field_defs
    ([@fields.length] + @fields.collect  do |name, type|
        case type
        when :integer
          0
        when :string
          1
        end
      end).pack('V*')
  end
  
  def read_header
    buffer = @io.read(@@header.length)
    throw IOError, "TupleStream header not found" unless buffer == @@header
    
    field_defs = create_field_defs
    buffer = @io.read(field_defs.length)
    throw IOError, "Different field spec in on-disk file" unless buffer = field_defs
  end
  
  def write_header
    @io.write(@@header)
    @io.write(create_field_defs)
  end
  
  def close(commit = true)
    @io.close
    @io = nil

    if @action == :write
      File.rename(@filename, @filename[0..@filename.length - '.wrk'.length - 1]) if commit
      File.delete(@filename) unless commit
    end
  end

  def closed?; @io.nil? end
  def eos?; @io.eof? end
end

if (__FILE__ == $0)
  class Links < TupleStream
    define_field :src_namespace, :integer
    define_field :src, :string
    define_field :dst_namespace, :integer
    define_field :dst, :string
    define_field :caption, :string
    define_field :type, :string
    define_field :weight, :integer
  
    define_comparator :by_src, [:src_namespace, :src]
    define_comparator :by_dst, [:dst_namespace, :dst]
  end
  
  #$= = true
  #  (0..7).each do |i|
  Links.open :full_name => "links_by_src", :order_by => :by_src do |l|
    (_,name,) = l.read
    until (item = l.read).nil?
      puts "#{item[1]} -> #{item[3]}"
      #      p item
      
      (_,name2,) = item
      if (name <=> name2) > 0
        puts "#{i}|||#{name}|||#{name2}"
        #        exit
      else
        name = name2
      end
      #      p item
    end
  end
  #  end
  
  
  exit
  #puts ["aeaWEFRawef"].pack('A*')

  #  Links.open('links2', :write) do |l|
  Links.open :name => :two, :mode => :write do |l|
    l.write(54, 23, 'sdfaewf')
    l.write(34554, 2343, 'dfgfgregerwf')
    l.write(54352, 346423, 'sdfaewretertertertret')
  end

  #  Links.open('links2', :read) do |l|
  #    p l.read
  #    p l.read
  #    p l.read
  #    p l.read
  #  end

  #exit

  class WikiPages < TupleStream
    define_field :namespace, :integer
    define_field :title, :string
    define_field :text, :string
  end

  class DummyStream < TupleStream
    define_field :value, :integer
  
    define_comparator :default, :value
  end
  
  DummyStream.open :mode => :write do |ds|
    (0..1000000).each { |i| ds.write(i) }
  end

  DummyStream.sort

  DummyStream.open :order => :default do |ds|
    (0..1000000).each { |i| (value, ) = ds.read; raise "#{i} != #{value}" unless value == i}
  end
  
  #
  #  DummyStream.open('dummy2', :read) do |ds|
  #    1000.times { |i| p ds.read }
  #    #  5000000.times { |i| v = ds.read; p v if i > 4999000 }
  #    #(0..5000000).each { |i| (value, ) = ds.read; raise "#{i} != #{value}" unless value == i}
  #  end

  #WikiPages.open('wikipages', :write) do |stream|
  ##  r = WikiPageReader.new('e:/wiki/all_pages.dat')
  #  wxp = WikiXmlParser.new('e:/Downloads/enwiki-latest-pages-articles.xml/enwiki-latest-pages-articles.xml')
  #  wxp.run do |page|
  #    stream.write(page[:namespace_key], page[:title], page[:text])
  #  end
  #end

  #p "ijerabcefdf".unpack('VA3V')
  #
  ##ts = TupleStream.define :integer, :integer, :string
  #timer :ts_write do
  #  #  Links.open :target => 'links' do |l|
  #  #    l.define2 :integer, :integer, :string
  #  #    5000000.times do |i|
  #  Links.open('links', :write) do |l|
  #    i = 0
  #    while i < 1000000
  #      l.write(i % 42, 1024 * (i % 12) + i, 'This is the caption!')
  #      i += 1
  #    end
  #  end
  #end
  #
  #timer :ts_sort do
  #  Links.sort('links', 'links2')
  #end
  #
  #exit
  #
  #timer :ts_read do
  #  Links.open('links', :read) do |l|
  #    while l.read
  #      
  #    end
  #  end
  #end
end
