require 'pp'
require 'pathname'
#require 'zlib'

input = "e:/wiki/all_pages.dat"

class WikiPageReader
  def initialize(filename)
    @filename = filename
  end
  
  def each(&block)
    #    Zlib::GzipReader.open(@filename) do |gz|
    index = 0
    File.open(@filename, 'rb') do |input|
      while !input.eof?# and index < 100000
        tuple = Marshal.load(input)
        if tuple.is_a?(Hash)
          tuple[:index] = index
          index += 1
        end
        block.call(tuple)
      end
    end
  end
  
  def emit(&block)
    each(&block)
  end
end

class FileStreams
  def initialize(root)
    @root = Pathname.new(root) if root
    @streams = Array.new
  end
  
  def open(filename, mode)
    filename = @root + filename if @root
    stream = File.open(filename, mode)
    @streams.push(stream)
    stream
  end
  
  def close
    @streams.each { |s| s.close }
  end
  
  def self.open(root = nil, &block)
    os = FileStreams.new(root)
    begin
      block.call(os)
    ensure
      os.close
    end
  end
end

def extract_pages
  page_count = {:category => 0, :template => 0, :regular => 0}

  FileStreams.open('e:/wiki') do |fs|
    templates = fs.open('templates.dat', 'w')
    categories = fs.open('categories.dat', 'w')
  
    reader = WikiPageReader.new(input)
    reader.each do |p|
      data = case p[:title]
      when /^Category:(.+)/
        page_count[:category] += 1
        {:stream => categories, :title => $1}
      when /^Template:(.+)/
        page_count[:template] += 1
        {:stream => templates, :title => $1}
      else
        page_count[:regular] += 1
        nil
      end
    
      Marshal.dump({:title => data[:title], :text => p[:text]}, data[:stream]) if data
    end
  end

  pp page_count
end
